#include <Keypad.h>
#include "tones.h"

const byte ROWS = 4; 
const byte COLS = 4; 

char hexaKeys[ROWS][COLS] = {
  {'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};

byte rowPins[ROWS] = {7, 6, 5, 4}; 
byte colPins[COLS] = {14, 15, 16, 17}; 

Keypad customKeypad = Keypad(makeKeymap(hexaKeys), rowPins, colPins, ROWS, COLS); 

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ST7735.h> // Hardware-specific library for ST7735
#include <SPI.h>

#if defined(ARDUINO_FEATHER_ESP32) // Feather Huzzah32
  #define TFT_CS         14
  #define TFT_RST        15
  #define TFT_DC         32

#elif defined(ESP8266)
  #define TFT_CS         4
  #define TFT_RST        16                                            
  #define TFT_DC         5

#else
  // For the breakout board, you can use any 2 or 3 pins.
  // These pins will also work for the 1.8" TFT shield.
  #define TFT_CS        10
  #define TFT_RST        8 // Or set to -1 and connect to Arduino RESET pin
  #define TFT_DC         9
#endif

Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS, TFT_DC, TFT_RST);

unsigned long _time, prev_millis;
uint8_t pos_x, pos_y;
uint8_t difficulty;

#define SOUND_PIN 2

uint8_t grid[1][9][9] = {
  {
    {2, 4, 6, 5, 7, 3, 8, 9, 1},  
    {5, 7, 3, 8, 1, 9, 2, 4, 6},  
    {8, 1, 9, 2, 4, 6, 5, 7, 3},  
    {7, 2, 8, 4, 5, 1, 3, 6, 9},  
    {3, 6, 5, 7, 9, 8, 4, 1, 2},  
    {4, 9, 1, 3, 6, 2, 7, 8, 5},  
    {9, 8, 2, 6, 3, 4, 1, 5, 7},  
    {1, 5, 4, 9, 2, 7, 6, 3, 8},  
    {6, 3, 7, 1, 8, 5, 9, 2, 4}
  },
};
uint8_t nr_grid = rand() % 1;
uint8_t sol[9][9] , puzzle[9][9];
uint8_t numberList[] = {1, 2, 3, 4, 5, 6, 7, 8, 9};
uint8_t counter = 0;
#define NOTE_SUSTAIN 100
void(* resetFunc) (void) = 0;

void setup(void) {
  

  

  Serial.begin(9600);
  tft.initR(INITR_BLACKTAB);      // Init ST7735S chip, black tab
  Serial.println(F("\nInitialized\n"));
  tft.fillScreen(ST77XX_BLACK);
  //delay(2000);
  prev_millis = millis();

  //Serial.println(time, DEC);

  tft.fillScreen(ST77XX_BLACK);
  tft.setRotation(2);

  uint8_t x = 90;
  while(!difficulty)
  {
    //
    tft.setCursor(12, 30);
    tft.setTextSize(3);
    tft.setTextColor(ST77XX_RED);
    tft.print("S");
    tft.setTextColor(ST77XX_ORANGE);
    tft.print("U");
    tft.setTextColor(ST77XX_YELLOW);
    tft.print("D");
    tft.setTextColor(ST77XX_GREEN);
    tft.print("O");
    tft.setTextColor(ST77XX_BLUE);
    tft.print("K");
    tft.setTextColor(ST77XX_MAGENTA);
    tft.print("U");
    //
    tft.setCursor(13, 65);
    tft.setTextSize(1);
    tft.setTextColor(ST77XX_WHITE);
    tft.print("Choose difficulty");
    //
    tft.setTextSize(2);
    tft.setTextColor(ST77XX_WHITE);
    tft.setCursor(40, 85);
    tft.print("EASY");
    tft.setCursor(40, 110);
    tft.print("HARD");
    //
    char key = customKeypad.getKey();
    if (key)
    {
      tft.fillCircle(33, x, 3, ST77XX_BLACK);
      if(key == 'A')
      {
        tone(SOUND_PIN,500,50);
        difficulty = x / 50;
      }
      if(key == 'C')
      {
        tone(SOUND_PIN,1000,30);
        x = 90;
      }
      if(key == 'D')
      {
        tone(SOUND_PIN,1000,30);
        x = 115;
      }
    }
    tft.fillCircle(33, x, 3, ST77XX_WHITE);
  }
  tft.fillScreen(ST77XX_BLACK);
  tft.setCursor(5, 65);
  tft.setTextSize(2);
  tft.setTextColor(ST77XX_WHITE);
  tft.print("Generating puzzle...");
  //
  genPuzzle();
  delay(1000);
  //
  
  //
  drawBackground();
  //
  tft.setRotation(2);
  tft.fillRect(0, 0, 70, 15, ST77XX_WHITE); 
  tft.setCursor(1, 1);
  tft.setTextColor(ST77XX_BLACK);
  tft.setTextSize(1);
  tft.print("TIME:00:00");
  //
  pos_x = pos_y = 4;
  tft.setRotation(1);
  tft.drawRect(1 + pos_x / 3 + 14 * pos_x, pos_y / 3 + 14 * pos_y, 13 , 13, ST77XX_BLUE);
  

  return;
}

void loop() {
  uint8_t frametime = millis() - prev_millis;
  prev_millis = millis();
  _time += frametime;
  //Serial.println(frametime, DEC);
  //Serial.println(time, DEC);
  if(_time / 1000 != (_time - frametime) / 1000) drawTime();
  drawSquare();
  drawPuzzle();
  uint8_t nr_empty = 0;
  for(uint8_t i = 0; i < 9; i++)
    for(uint8_t j = 0 ; j < 9; j++)
      if(puzzle[i][j] == 0) nr_empty++;
  Serial.println(nr_empty);
  if(nr_empty == 0)
  {
    // Board is filled
    uint8_t ok = 1;
    for(uint8_t i = 0; i < 9; i++)
      for(uint8_t j = 0 ; j < 9; j++)
        if(puzzle[i][j] != sol[i][j]) ok = 0;

    if(ok)
    {
      // Game is won
      tft.fillScreen(ST77XX_BLACK);
      tft.setCursor(12, 30);
      tft.setTextSize(3);
      tft.setTextColor(ST77XX_RED);
      tft.print("S");
      tft.setTextColor(ST77XX_ORANGE);
      tft.print("U");
      tft.setTextColor(ST77XX_YELLOW);
      tft.print("D");
      tft.setTextColor(ST77XX_GREEN);
      tft.print("O");
      tft.setTextColor(ST77XX_BLUE);
      tft.print("K");
      tft.setTextColor(ST77XX_MAGENTA);
      tft.print("U");
      tft.setCursor(1, 80);
      tft.setTextSize(3);
      tft.setTextColor(ST77XX_WHITE);
      tft.print("YOU WIN!");
      for(uint8_t i = 0; i < 3; i++)
      {
        tone(SOUND_PIN,NOTE_A5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_B5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_C5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_B5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_C5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_D5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_C5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_D5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_E5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_D5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_E5);
        delay(NOTE_SUSTAIN);
        tone(SOUND_PIN,NOTE_E5);
        delay(3 * NOTE_SUSTAIN);
      }
      noTone(SOUND_PIN);
      delay(1000);
      while(1)
      {
        char key = customKeypad.getKey();
        if(key == 'A')
        {
          resetFunc();
        }
      }
    }
  }
  
}

void drawPuzzle()
{
  tft.setRotation(2);
  tft.setTextColor(ST77XX_BLACK);
  tft.setTextSize(1);
  for(uint8_t i = 0; i < 9; ++i)
  {
    for(uint8_t j = 0; j < 9; ++j)
    {
      if(puzzle[i][j] != 0)
      {
        tft.setTextColor(ST77XX_BLACK);
        if(puzzle[i][j] != grid[nr_grid][i][j])
        {
          tft.setTextColor(ST77XX_BLUE);
          if(difficulty == 1 && puzzle[i][j] != sol[i][j])
          {
            tft.setTextColor(ST77XX_RED);
          }
        }
        
        tft.setCursor(4 + i / 3 + 14 * i, 149 - (j / 3 + 14 * j));    
        tft.print(puzzle[i][j]);
      }
      
    }
  }
  
}
void inputNumber(char key)
{
  if (key)
  {
    tone(SOUND_PIN,500,50);
    Serial.println(key - 48);
    if(grid[nr_grid][pos_y][pos_x] == 0)
    {
      puzzle[pos_y][pos_x] = key - 48;
      tft.fillRect(2 + pos_x / 3 + 14 * pos_x, 1 + pos_y / 3 + 14 * pos_y, 11 , 11, ST77XX_WHITE);
      if(key == '0') return;
      tft.setRotation(2);
      tft.setTextColor(ST77XX_BLACK);
      tft.setTextSize(1);
      tft.setCursor(4 + pos_y / 3 + 14 * pos_y, 149 - (pos_x / 3 + 14 * pos_x));
      tft.print(puzzle[pos_y][pos_x]);
    }
  }
}
bool getInput()
{
  char key = customKeypad.getKey();
  if (key)
  {
    switch (key)
    {
      case '*':
        pos_y = max(pos_y - 1, 0);
        tone(SOUND_PIN,1000,30);
        break;
      case '#':
        pos_y = min(pos_y + 1, 8);
        tone(SOUND_PIN,1000,30);
        break;
      case 'D':
        pos_x = max(pos_x - 1, 0);
        tone(SOUND_PIN,1000,30);
        break;
      case 'C':
        pos_x = min(pos_x + 1, 8);
        tone(SOUND_PIN,1000,30);
        break;
      case 'A':
        break;
      case 'B':
        break;
      default:
        inputNumber(key);
        return 0;
    }
    Serial.println(key);
    return 1;
  }
  return 0;
}
void drawSquare()
{
  tft.setRotation(1);
  uint8_t prev_x = pos_x, prev_y = pos_y;
  if(getInput())
  {
    tft.drawRect(1 + prev_x / 3 + 14 * prev_x, prev_y / 3 + 14 * prev_y, 13 , 13, ST77XX_WHITE);
    tft.drawRect(1 + pos_x / 3 + 14 * pos_x, pos_y / 3 + 14 * pos_y, 13 , 13, ST77XX_BLUE);
  }
}
void drawBackground()
{
  tft.setRotation(1);
  tft.fillScreen(ST77XX_WHITE);
  //tft.drawLine(10,1,10,100,ST77XX_BLACK);
  int i = 0, j = 0, offset = 1;
  for(i = 0; i < 3; i++)
  {
    for(j = 0; j <= 3; j++)
    {
      tft.drawLine(i * 43 + j * 14, 0, i * 43 + j * 14, 128, ST77XX_BLACK);
      tft.drawLine(0, i * 43 + j * 14 - offset, 128, i * 43 + j * 14 - offset, ST77XX_BLACK);
    }
  }
  tft.drawLine(129, 0, 129, 128, ST77XX_BLACK);
}


void drawTime()
{
  tft.setRotation(2);
  //drawBackground();
  tft.fillRect(0, 0, 70, 15, ST77XX_WHITE); 
  tft.setTextWrap(false);
  //tft.fillScreen(ST77XX_BLACK);
  tft.setCursor(1, 1);
  tft.setTextColor(ST77XX_BLACK);
  tft.setTextSize(1);
  tft.print("TIME:");
  int seconds = _time / 1000;
  tft.print(seconds / 600);
  tft.print(seconds % 600 / 60);
  tft.print(":");
  tft.print(seconds % 60 / 10);
  tft.print(seconds % 10);
}


// Game logic

uint8_t solveGrid()
{
    uint8_t row, col, i, j, k, value, ok, idx;
    for(i = 0; i < 81; ++i)
    {
     //Serial.println(i);
     row = i / 9;
     col = i % 9;
     if(puzzle[row][col] == 0)
     {
       for(value = 1; value < 10; ++value)
       {
         // Check that this value has not already be used on this row
         ok = 1;
         for(j = 0; j < 9 && ok; ++j)
         {
           if(puzzle[row][j] == value)
             ok = 0;
         }
         for(j = 0; j < 9 && ok; ++j)
         {
           if(puzzle[j][col] == value)
             ok = 0;
         }
         if(ok)
         {
           // Identify which of the 9 squares we are working on and
           // check that this value has not already be used on this 3x3 square
           
           ok = 0;
           
           for(j = (i / 9) / 3 * 3; j < (i / 9) / 3 * 3 + 3 + 3 && !ok; ++j)
           {
             for(k = (i % 9) / 3 * 3; k < (i % 9) / 3 * 3 + 3 && !ok; ++k)
             {
               if(value == puzzle[j][k])
                 ok = 1;
             }
           }
           if(!ok)
           {
             puzzle[row][col] = value;
             ok = 1;
             for(j = 0; j < 9; ++j)
              for(k = 0; k < 9; ++k)
               if(puzzle[j][k] == 0)
               {
                  ok = 0;
                  break;
               }
             if(ok)
             {
               counter++;
               break;
             }
             else
             {
               if(solveGrid())
                 return 1;
             }
           }
         }
       }
       break;
     }
    }
    puzzle[row][col] = 0;
    return 0;
}


void shuffle()
{
    srand(random(analogRead(A5)));
    uint8_t nr_shuffles = rand() % 10 + 10;// should be enough
    uint8_t a, b, aux, i;
    for(i = 0; i < nr_shuffles; ++i)
    {
     do
     {
       a = rand() % 9;
       b = rand() % 9;
     } while(a == b);
     aux = numberList[a];
     numberList[a] = numberList[b];
     numberList[b] = aux;
    }
}

int freeRam () {
  extern int __heap_start, *__brkval;
  int v;
  return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}

void fillGrid(uint8_t nr_grid)
{
    shuffle();
    uint8_t i, j;
    for(i = 0; i < 9; i++)
      Serial.println(numberList[i]);
    for(i = 0; i < 9; i++)
      for(j = 0; j < 9 ; j++)
      {
        grid[nr_grid][i][j] = numberList[grid[nr_grid][i][j] - 1];
      }
      
}


void genPuzzle()
{
    
    Serial.println(F("Generating solved grid..."));
    fillGrid(nr_grid);
    Serial.println(F("Solved grid done"));
    //return;
    for(uint8_t i = 0; i < 9; ++i)
       for(uint8_t j = 0; j < 9; ++j)
       {
         sol[i][j] = grid[nr_grid][i][j];
       }
     
    uint8_t nr_remove = 0;// 30 45
    if(difficulty == 1) nr_remove = 30;
    if(difficulty == 2) nr_remove = 45;
    uint8_t row, col, backup;
    counter = 1;
    Serial.println(F("Generating puzzle..."));
    while(nr_remove > 0)
    {
     // Select a random cell that is not already empty
     row = rand() % 9;
     col = rand() % 9;
     while(grid[nr_grid][row][col] == 0)
     {
       row = rand() % 9;
       col = rand() % 9;
     }
     // Remember its cell value in case we need to put it back
     backup = grid[nr_grid][row][col];
     grid[nr_grid][row][col] = 0;
     // Make a full copy of the grid
     for(uint8_t i = 0; i < 9; ++i)
       for(uint8_t j = 0; j < 9; ++j)
         puzzle[i][j] = grid[nr_grid][i][j];
     // Count the number of solutions
     counter = 0;
     solveGrid();
     if(counter != 1)
     {
       grid[nr_grid][row][col] = backup;
     }
     else nr_remove--;
    }
    Serial.println(F("Puzzle generated..."));

}
